/*  sound.c
 *
 *  Soundcard handling functions of xpsk31 application
 */

/*
 *  xpsk31: An application to transmit and receive
 *  PSK31 signals using a computer's sound card
 *
 *  Copyright (C) 2002  Neoklis Kyriazis
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */

#include "xpsk31.h"

int
  dsp_in_fd  = 0, /* File descriptor of DSP device   */
  dsp_out_fd = 0, /* File descriptor of DSP device   */
  mix_fd     = 0; /* File descriptor of mixer device */

/* Receive freq error for freq ctrl */
extern int freq_error;

int
  num_chan,		/* Stereo(=2) or Mono(=1) selection */
  chan_used,	/* Channel in use: left=0, right=1; */
  xmit_buf_size,/* Xmit DSP signal samples buffer   */
  recv_buf_size,/* Recv DSP signal samples buffer   */
  recv_buf_idx;	/* Index to Rx signal samples buffer*/

/* Signal samples buffers */
unsigned char *xmit_buffer;
unsigned char *recv_buffer;

/* Config data */
extern rc_data_t rc_data;

/* fft in/out buffers */
extern int
  *fft_in_r,
  *fft_out_r,
  *fft_out_i;

/*------------------------------------------------------------------------*/

/*  Setup_Sound_Card()
 *
 *  Sets up mixer and DSP devices
 */

  void
Setup_Sound_Card( void )
{
  int
	new_recsrc,  /* New recording source (RECORD_SRC) */
	level,       /* Rec/Volume level request to mixer */
	itmp;

  /* Mixer device to be set-up (input or reclev) */
  unsigned int device;

  /* Error messages */
  char message[85];

  /* Sound device names from souncard.h */
  char *dev_name[SOUND_MIXER_NRDEVICES] = SOUND_DEVICE_NAMES;

  /* File lock information */
  struct flock lockinfo;


  /* Return if sound card is set-up */
  if( isFlagSet(SOUND_CARD_SETUP) )
	return;

  /*** Open and lock mixer device ***/
  /* Open mixer device, abort on error */
  if( (mix_fd = open(rc_data.mixer, O_RDWR, 0)) == -1 )
  {
	perror( rc_data.mixer );
	snprintf( message, 85,
		"Unable to open %s\n"
		"Quit xpsk31 and correct", rc_data.mixer);
	Error_Dialog( message, TRUE );
	return;
  }

  /* Attempt to lock entire mixer device file */
  lockinfo.l_type   = F_WRLCK;
  lockinfo.l_whence = SEEK_SET;
  lockinfo.l_start  = 0;
  lockinfo.l_len    = 0;

  /* If mixer device is already locked, abort */
  if( fcntl( mix_fd, F_SETLK, &lockinfo ) < 0 )
  {
	perror( rc_data.mixer );
	fcntl( mix_fd, F_GETLK, &lockinfo );
	fprintf( stderr, "xpsk31: Lock %s: Device is locked by pid %d\n",
		rc_data.mixer, lockinfo.l_pid );
	snprintf( message, 85,
		"Unable to lock %s\n"
		"Device locked by pid %d\n"
		"Quit xpsk31 and correct",
		rc_data.mixer, lockinfo.l_pid );
	Error_Dialog( message, TRUE );
	return;
  }

  /*** Setup mixer parameters ***/
  /* Find number of recording source matching source name */
  for( new_recsrc = 0; new_recsrc < SOUND_MIXER_NRDEVICES; new_recsrc++ )
	if( strcmp( dev_name[new_recsrc], rc_data.rec_src ) == 0 )
	  break;

  /* Make recording source mask */
  itmp = 1 << new_recsrc;

  /* Attempt to select recording source, abort on error */
  if( ioctl(mix_fd, SOUND_MIXER_WRITE_RECSRC, &itmp) == -1 )
  {
	perror("xpsk31: SOUND_MIXER_WRITE_RECSRC");
	snprintf( message, 85,
		"Failed to select Recording\n"
		"Source: %s\n"
		"Quit xpsk31 and correct", rc_data.rec_src );
	Error_Dialog( message, TRUE );
	return;
  }
  if( itmp != 1 << new_recsrc )
  {
	snprintf( message, 85,
		"Failed to select Recording\n"
		"Source: %s\n"
		"Quit xpsk31 and correct", rc_data.rec_src );
	Error_Dialog( message, TRUE );
	return;
  }

  /* Zero line level to silence (my) speakers */
  level = itmp = 0;
  if( ioctl(mix_fd, MIXER_WRITE(new_recsrc), &level) == -1 )
  {
	perror("xpsk31: Setting Recording source level");
	snprintf( message, 85,
		"Failed to set %s level\n"
		"Quit xpsk31 and correct", rc_data.rec_src );
	Error_Dialog( message, TRUE );
	return;
  }
  if( level != itmp )
  {
	snprintf( message, 85,
		"Failed to set %s level\n"
		"Quit xpsk31 and correct", rc_data.rec_src );
	Error_Dialog( message, TRUE );
	return;
  }

  /* Find recording level device */
  for( device = 0; device < SOUND_MIXER_NRDEVICES; device++ )
	if( strcmp( dev_name[device], rc_data.lev_dev ) == 0 )
	  break;

  if( device ==  SOUND_MIXER_NRDEVICES )
  {
	snprintf( message, 85,
		"Failed to select Recording\n"
		"Level device: %s\n"
		"Quit xpsk31 and correct", rc_data.lev_dev );
	Error_Dialog( message, TRUE );
	return;
  }

  /* Write recording level to mixer, abort on error */
  level = itmp = rc_data.rec_lev | (rc_data.rec_lev << 8);
  if( ioctl(mix_fd, MIXER_WRITE(device), &level) == -1 )
  {
	perror("xpsk31: Setting Recording level");
	Error_Dialog(
		"Failed to set recording level\n"
		"Quit xpsk31 and correct", TRUE );
	return;
  }
  if( level != itmp )
  {
	Error_Dialog(
		"Failed to set Recording level\n"
		"Quit xpsk31 and correct", TRUE );
	return;
  }

  /* Write level to Volume, abort on error */
  level = itmp = rc_data.vol_lev | (rc_data.vol_lev << 8);
  if( ioctl(mix_fd, SOUND_MIXER_WRITE_VOLUME, &level) == -1 )
  {
	perror("xpsk31: Setting Volume level");
	Error_Dialog(
		"Failed to set Volume level\n"
		"Quit xpsk31 and correct", TRUE );
	return;
  }
  if( level != itmp )
  {
	Error_Dialog(
		"Failed to set Volume level\n"
		"Quit xpsk31 and correct", TRUE );
	return;
  }

  /* Write level to PCM, abort on error */
  level = itmp = rc_data.vol_lev | (rc_data.vol_lev << 8);
  if( ioctl(mix_fd, SOUND_MIXER_WRITE_PCM, &level) == -1 )
  {
	perror("xpsk31: Setting PCM level");
	Error_Dialog(
		"Failed to set PCM level\n"
		"Quit xpsk31 and correct", TRUE );
	return;
  }
  if( level != itmp )
  {
	Error_Dialog(
		"Failed to set PCM level\n"
		"Quit xpsk31 and correct", TRUE );
	return;
  }

  /*** Open and setup dsp device(s) ***/
  /* If only one device specified, open input dsp r/w */
  if( strcmp(rc_data.dsp_in, rc_data.dsp_out) == 0 )
  {
	if( !Setup_Dsp_Dev(rc_data.dsp_in, O_RDWR, &dsp_in_fd) )
	  return;
	dsp_out_fd = dsp_in_fd;
  }
  else /* Setup separate dsp in/out */
  {
	if( !Setup_Dsp_Dev(rc_data.dsp_in, O_RDONLY, &dsp_in_fd) )
	  return;

	if( !Setup_Dsp_Dev(rc_data.dsp_out, O_WRONLY, &dsp_out_fd) )
	  return;
  }

  SetFlag( SOUND_CARD_SETUP );

} /* End of Setup_Sound_Card() */

/*------------------------------------------------------------------------*/

/*  Setup_Dsp_Dev()
 *
 *  Sets up a given dsp device
 */

  gboolean
Setup_Dsp_Dev( char *dsp_dev, int mode, int *dsp_fd )
{
  int
	dsp_speed,   /* DSP sampling speed as in xpsk31.h */
	frag_size,   /* DSP buffer (fragment) size in kb  */
	sample_fmt,  /* DSP sample format, 8-bit unsigned */
	itmp;

  /* Error messages */
  char message[85];

  /* File lock information   */
  struct flock lockinfo;


  /* Open input DSP device, abort on error */
  if( (*dsp_fd = open(dsp_dev, mode, 0)) == -1 )
  {
	perror( dsp_dev );
	snprintf( message, 85,
		"Unable to open %s\n"
		"Quit xpsk31 and correct", dsp_dev);
	Error_Dialog( message, TRUE );
	return FALSE;
  }

  /* Attempt to lock entire DSP device file */
  if( mode == O_RDONLY )
  {
	lockinfo.l_type   = F_RDLCK;
	lockinfo.l_whence = SEEK_SET;
	lockinfo.l_start  = 0;
	lockinfo.l_len    = 0;
  }
  else
	if( mode == O_WRONLY )
	{
	  lockinfo.l_type   = F_WRLCK;
	  lockinfo.l_whence = SEEK_SET;
	  lockinfo.l_start  = 0;
	  lockinfo.l_len    = 0;
	}
	else
	{
	  lockinfo.l_type   = F_RDLCK | F_WRLCK;
	  lockinfo.l_whence = SEEK_SET;
	  lockinfo.l_start  = 0;
	  lockinfo.l_len    = 0;
	}


  /* If DSP device is already locked, abort */
  if( fcntl( *dsp_fd, F_SETLK, &lockinfo ) < 0 )
  {
	perror( dsp_dev );
	fcntl( *dsp_fd, F_GETLK, &lockinfo );
	fprintf( stderr, "xpsk31: Lock %s: Device is locked by pid %d\n",
		dsp_dev, lockinfo.l_pid );
	snprintf( message, 85,
		"Unable to lock %s\n"
		"Device locked by pid %d\n"
		"Quit xpsk31 and correct",
		dsp_dev, lockinfo.l_pid );
	Error_Dialog( message, TRUE );
	return FALSE;
  }

  /* Set fragment size according to stereo/mono mode */
  itmp = 1; frag_size = 0;
  while( (itmp << frag_size) < rc_data.psk_elem ) frag_size++;
  frag_size += num_chan << 18;
  ioctl( *dsp_fd, SNDCTL_DSP_SETFRAGMENT, &frag_size );

  /* Set sample format to unsigned 8-bit, abort on error */
  sample_fmt = AFMT_U8;
  if( ioctl(*dsp_fd, SNDCTL_DSP_SETFMT, &sample_fmt) == -1 )
  {
	perror( "xpsk31: SNDCTL_DSP_SETFMT");
	Error_Dialog(
		"Failed to set dsp sample\n"
		"format: AFMT_U8\n"
		"Quit xpsk31 and correct", TRUE );
	return FALSE;
  }
  if( sample_fmt != AFMT_U8 )
  {
	Error_Dialog(
		"Failed to set dsp sample\n"
		"format: AFMT_U8\n"
		"Quit xpsk31 and correct", TRUE );
	return FALSE;
  }

  /* Set stereo/mono mode, continue on error */
  itmp = num_chan;
  if( ioctl(*dsp_fd, SNDCTL_DSP_CHANNELS, &itmp) == -1 )
	perror("SNDCTL_DSP_CHANNELS");

  /* Setup DSP sampling speed, abort on error */
  dsp_speed = DSP_SPEED;
  if( ioctl(*dsp_fd, SNDCTL_DSP_SPEED, &dsp_speed) == -1 )
  {
	perror( "xpsk31: SNDCTL_DSP_SPEED");
	snprintf( message, 85,
		"Failed to set DSP\n"
		"speed to %d\n"
		"Quit xpsk31 and correct", DSP_SPEED );
	Error_Dialog( message, TRUE );
	return FALSE;
  }
  if( dsp_speed != DSP_SPEED )
  {
	snprintf( message, 85,
		"Failed to set DSP\n"
		"speed to %d\n"
		"Quit xpsk31 and correct", DSP_SPEED );
	Error_Dialog( message, TRUE );
	return FALSE;
  }

  return TRUE;

} /* Setup_Dsp_Dev */

/*------------------------------------------------------------------------*/

/*  Signal_Sample()
 *
 *  Gets the next DSP sample of the signal input.
 */

  int
Signal_Sample( void )
{
  int
	sample_val, /* Signal sample value */
	sample_cnt; /* Number of samples read from DSP */

  static int
	fft_idx = 0, /* fft input buffer idx */
			cnt = 0;	 /* Count of calls to this function */

  /* Counts num of samples after which a sample *
   * is drop or repeated to correct freq error  */
  static int xtra_sample_cnt = 0;

  /* Refill recv DSP samples buffer when needed */
  if( recv_buf_idx >= recv_buf_size )
  {
	/* Start buffer index according to stereo/mono mode */
	if( num_chan == 2 ) /* Stereo */
	  recv_buf_idx = chan_used;
	else
	  recv_buf_idx = 0;

	/* Read audio samples from DSP, abort on error */
	if( (sample_cnt = read(dsp_in_fd, recv_buffer, recv_buf_size)) == -1 )
	{
	  perror( "xpsk31: dsp read()" );
	  Error_Dialog(
		  "Error reading from dsp device\n"
		  "Quit xpsk31 and correct", TRUE );
	  SetFlag( DSP_IO_ERROR );
	}
	if( sample_cnt != recv_buf_size )
	{
	  Error_Dialog(
		  "Error reading from dsp device\n"
		  "Quit xpsk31 and correct", TRUE );
	  SetFlag( DSP_IO_ERROR );
	}

  } /* End of if( buffer_idx == buffer_size ) */

  /* Remove neutral value (0x80) from samples */
  sample_val = recv_buffer[recv_buf_idx] - 0x80;

  /* Return extra sample if freq error positive */
  if( (freq_error > 0 ) &&
	  (xtra_sample_cnt++ >= AUDIO_FREQx10/freq_error) )
  {
	xtra_sample_cnt = 0;
	return( sample_val );
  }

  /* Drop a sample if freq error negative */
  if( (freq_error < 0 ) &&
	  (xtra_sample_cnt++ >= -AUDIO_FREQx10/freq_error) )
  {
	xtra_sample_cnt = 0;
	recv_buf_idx += num_chan;
  }

  /* Increment according to mono/stereo mode */
  recv_buf_idx += num_chan;

  /* Display waterfall when input buffer full */
  /* fft_stride samples are added for each input element */
  if( isFlagClear(ENABLE_MAGNIPHASE) )
  {
	if( cnt == 0 )
	  fft_in_r[fft_idx] = sample_val;
	else
	  fft_in_r[fft_idx] += sample_val;

	if( ++cnt == rc_data.fft_stride )
	{
	  fft_idx++;
	  cnt = 0;
	}

	if( fft_idx >= FFT_SIZE )
	{
	  fft_idx = 0;
	  Display_Waterfall();
	}

  } /* if( isFlagClear(ENABLE_MAGNIPHASE) ) */

  return( sample_val );

} /* End of Signal_Sample() */

/*------------------------------------------------------------------------*/

/*  DSP_Write()
 *
 *  Write a samples buffer to DSP
 */

  void
DSP_Write( unsigned char *buffer, int buff_size )
{
  int sample_cnt;

  /*** Write samples buffer to DSP or file ***/
  if( (sample_cnt = write(dsp_out_fd, buffer, buff_size)) == -1 )
  {
	perror( "xpsk31: dsp read()" );
	Error_Dialog(
		"Error writing to dsp device\n"
		"Quit xpsk31 and correct", TRUE );
	SetFlag( DSP_IO_ERROR );
  }
  if( sample_cnt != buff_size )
  {
	Error_Dialog(
		"Error writing to dsp device\n"
		"Quit xpsk31 and correct", TRUE );
	SetFlag( DSP_IO_ERROR );
  }

} /* DSP_Write() */

/*------------------------------------------------------------------------*/
